#include <rt/context.h>
#include <rt/cycle.h>
#include <rt/idle.h>
#include <rt/interrupt.h>
#include <rt/log.h>
#include <rt/mpu.h>
#include <rt/syscall.h>
#include <rt/task.h>
#include <rt/trap.h>

#include <stdbool.h>
#include <stdint.h>

#define PROFILE_R (__ARM_ARCH_PROFILE == 'R')
#define PROFILE_M (__ARM_ARCH_PROFILE == 'M')

#define V6M ((__ARM_ARCH == 6) && PROFILE_M)
#define V8M ((__ARM_ARCH == 8) && PROFILE_M)

#ifdef __ARM_FP
#define FPU 1
#else
#define FPU 0
#endif

#if PROFILE_R
#include "r/r.c"
#elif PROFILE_M
#include "m/m.c"
#else
#error "unknown __ARM_ARCH_PROFILE"
#endif

__attribute__((noreturn)) void rt_task_entry(void (*fn)(void));

__attribute__((noreturn)) void rt_task_entry_arg(uintptr_t arg,
                                                 void (*fn)(uintptr_t));

void *rt_context_init(void (*fn)(void), void *stack, size_t stack_size)
{
    struct context *const ctx = context_init(stack, stack_size, (uintptr_t)fn);
    ctx->pc = (uint32_t)rt_task_entry;
    ctx->r0 = (uint32_t)fn;
    return ctx;
}

void *rt_context_init_arg(void (*fn)(uintptr_t), uintptr_t arg, void *stack,
                          size_t stack_size)
{
    struct context *const ctx = context_init(stack, stack_size, (uintptr_t)fn);
    ctx->pc = (uint32_t)rt_task_entry_arg;
    ctx->r0 = arg;
    ctx->r1 = (uint32_t)fn;
    return ctx;
}

__attribute__((noreturn, weak)) void rt_idle(void)
{
    rt_task_drop_privilege();
    for (;;)
    {
        __asm__("wfi" :::);
    }
}

__attribute__((noreturn, weak)) void rt_trap(void)
{
    for (;;)
    {
        __asm__("bkpt" :::);
    }
}

void rt_syscall(void)
{
    __asm__("svc 0" ::: "memory");
}

#if V6M
#include "m/atomic-v6.c"
#endif
