#pragma once

@ vim:ft=arm

/* This file provides the entry and exit sequences for the syscall handler
 * specific to the Vectored Interrupt Manager (VIM) and the Software
 * IRQ (SW_IRQ), as implemented on the TI Sitara family (AM26x). */

#include "vic.h"
#include "../mode.h"

.macro vic_syscall_start
    /* Clear the SW_IRQ. The VIM automatically masks all interrupts of the same
     * priority including SW_IRQ until we store a value into IRQVEC. */
    ldr r0, =RT_VIC_SW_IRQ_GROUP
    mov r1, RT_VIC_SW_IRQ_BITMASK
    str r1, [r0, 4]
.endm

.macro vic_syscall_finish
    // Clear the mask and re-evaluate pending interrupts.
    ldr r0, =RT_VIC_VIM_IRQVEC
    str r0, [r0]
.endm

/* To keep the VIM management code for the IRQ simpler, we provide a separate
 * SVC handler that just triggers the SW IRQ and returns.
 * An alternative would be to set the VIM IRQ priority mask in
 * vic_syscall_start and clear it in vic_syscall_finish. This step is only
 * required when handling an SVC, but it would affect syscall IRQs as well. */

.macro vic_syscall_pend
    push {r0, r1}
    ldr r0, =RT_VIC_SW_IRQ_GROUP
    mov r1, RT_VIC_SW_IRQ_BITMASK
    str r1, [r0]
    /* Because this is invoked from the SVC handler for rt_syscall, we must
     * ensure that the SW IRQ store completes before returning from the SVC
     * handler. Returning from the SVC handler synchronizes instructions, so
     * only a dsb is needed here. */
    dsb
    pop {r0, r1}
.endm

    .syntax unified
#ifdef __thumb__
    .thumb
#else
    .arm
#endif

    .section .text.rt_syscall_handler_svc,"ax",%progbits
    .global rt_syscall_handler_svc
    .type rt_syscall_handler_svc, %function
rt_syscall_handler_svc:
    vic_syscall_pend
    subs pc, lr, 0

    .size rt_syscall_handler_svc, .-rt_syscall_handler_svc


/* At the cost of 2 extra instructions and 16 bytes of per-task stack space,
 * the SVC handler can re-enable interrupts to minimize the latency of other
 * interrupts. An application can choose either depending on system
 * requirements. */
    .section .text.rt_syscall_handler_svc_ie,"ax",%progbits
    .global rt_syscall_handler_svc_ie
    .type rt_syscall_handler_svc_ie, %function
rt_syscall_handler_svc_ie:
    srsdb sp!, MODE_SYS
    cpsie i, MODE_SYS
    vic_syscall_pend
    rfeia sp!

    .size rt_syscall_handler_svc_ie, .-rt_syscall_handler_svc_ie
