#pragma once

#include <rt/list.h>
#include <rt/sem.h>

#ifdef __cplusplus
extern "C" {
#endif

struct rt_mutex;

void rt_mutex_init(struct rt_mutex *mutex);

void rt_mutex_lock(struct rt_mutex *mutex);

void rt_mutex_unlock(struct rt_mutex *mutex);

bool rt_mutex_trylock(struct rt_mutex *mutex);

bool rt_mutex_timedlock(struct rt_mutex *mutex, unsigned long ticks);

struct rt_mutex
{
    rt_atomic_uintptr_t holder;
    struct rt_list wait_list;
    struct rt_list list;
    int level;
};

#define RT_MUTEX_INIT_COMMON(name, level_)                                     \
    {                                                                          \
        .holder = (uintptr_t)0, .wait_list = RT_LIST_INIT(name.wait_list),     \
        .list = RT_LIST_INIT(name.list), .level = (level_),                    \
    }

#define RT_MUTEX_INIT(name) RT_MUTEX_INIT_COMMON(name, -1)

#define RT_MUTEX_INIT_RECURSIVE(name) RT_MUTEX_INIT_COMMON(name, 0)

#define RT_MUTEX(name) struct rt_mutex name = RT_MUTEX_INIT(name)

#define RT_MUTEX_RECURSIVE(name)                                               \
    struct rt_mutex name = RT_MUTEX_INIT_RECURSIVE(name)

#define RT_MUTEX_WAITED_MASK ((uintptr_t)1)

#define RT_MUTEX_HOLDER_MASK (~RT_MUTEX_WAITED_MASK)

/* Try to lock the mutex from a given task.
 * This function should not be called directly. */
bool rt_mutex_trylock_with_task(struct rt_mutex *mutex, struct rt_task *task);

static inline struct rt_task *rt_mutex_holder(const struct rt_mutex *mutex)
{
    return (struct rt_task *)(rt_atomic_load(&mutex->holder,
                                             RT_ATOMIC_RELAXED) &
                              RT_MUTEX_HOLDER_MASK);
}

#ifdef __cplusplus
}
#endif
