use core::{
    future::Future,
    pin::pin,
    task::{Context, Poll, RawWaker, RawWakerVTable, Waker},
};

use crate::{semaphore_binary, sync::Semaphore};

static SEM_WAKER_VTABLE: RawWakerVTable = RawWakerVTable::new(clone, wake, wake_by_ref, drop);

semaphore_binary!(SEM_WAKER);

unsafe fn clone(w: *const ()) -> RawWaker {
    RawWaker::new(w, &SEM_WAKER_VTABLE)
}

unsafe fn wake(w: *const ()) {
    wake_by_ref(w)
}

unsafe fn wake_by_ref(w: *const ()) {
    let sem = &*(w as *mut () as *mut Semaphore);
    sem.post();
}

unsafe fn drop(_: *const ()) {}

pub fn block_on<F: Future>(f: F) -> F::Output {
    let waker = unsafe {
        Waker::from_raw(RawWaker::new(
            &SEM_WAKER as *const Semaphore as *const (),
            &SEM_WAKER_VTABLE,
        ))
    };
    let mut cx = Context::from_waker(&waker);
    let mut f = pin!(f);
    loop {
        if let Poll::Ready(o) = f.as_mut().poll(&mut cx) {
            return o;
        }
        SEM_WAKER.wait();
    }
}
